unit mStrGrid;

{Version 3.0a beta8,  24.6.1997,  PostCardWare, Albrecht Mengel, mengel@stat-econ.uni-kiel.de
 See mStrGrid.txt

Problem: Eigene OnDraw-Ereignisse sollen sich vernnftig anpassen.
Alt: Eigene OnDraw werden berschrieben durch interne. Neu: Eigene sollten nur Textfarbe setzen (oder?)

Bugs known (in work at...) Freezed Range Selection does not show new selection
Bugs fixed: Frozen Selection remains red while standard selection keeps invisible

If you have problems/new ideas with mStrGrid or mStrList, please feel free to email me!

New: Selecting
     Properties SelectMethod = (smRange,smCompleteRows,smFree);
                Selected[ACol,ARow:Integer]:boolean   Here you can set or test selections of any cell
     Method     UnSelect    Sets selection to actual cell/range
     Property   FreezeSelection:Boolean
                Here any manual selection can be disabled

New: Clicking on fixed cells sorts
     Property ClickSortCols:boolean   enables/disables calling of clicking fixed cols event
     Property ClickSortRows:boolean   enables/disables calling of clicking fixed rows event
     Event    OnFixedColClick  called if clicked on fixed col (and ClickSortCols=true)
     Event    OnFixedRowClick  called if clicked on fixed row (and ClickSortRows=true)

New: Assign  Assigns new contents to the grid. Accepts any TStringGrid, TStrings, TmStrGrid.

New: Method ReorderCompleteRows(KeyCol:integer; OrderValues:TStrings);
         Just call with the wished order of the KeyColumn as StringList.
         This sets all columns into that same order.
         Known Problems: No duplicate values allowed in OrderValues nor in the key column
                         OderValues must be a true permutation of the key column

New: Multi-Key sorting is possible now with SortCompletRows (and SortCompleteColumns):
     First sort the least important column(s), then sort the most important column(s).
     (The order of equal-valued entries is kept).
     As result, the whole table is sorted after the most important column. If any values there
     are equal, the table is sorted by the next less important value, and so on.
     (There is an expample supplied to test this property)

Tip for coloring mStrGrid: see mStrGrid.txt

New date sorting Properties: DateTimeFormat:string; ShortYearExpand:boolean; ShortYearBorder:(0..99)
     When sorting dates, they are interpreted in the kind as DateTimeFormat is set to.
     Default is the value of (the global) ShortDateFormat+' '+LongTimeFormat.
     See mStrGrid.txt for more description

For ever: Unsolvable problem with Click into fixed cells: The (programmer of the) OnClick procedure
whishes to use COL and ROW to determine which cell was clicked. But the assigment of COL
and ROW to a fixed cell does not work (I cannot change that). To implement the wished behavior
I set temporarily fixedcols/fixedrows to zero, while the OnClick is running.
The published methods of this unit know the original values and work correct. But
You as normal programmer find fixedcols/fixedrows in OnClick to be zero, when clicked a
fixed cell. As cold comfort you may read the properties SavedFixedRows and SavedFixedCols,
which contain the original values (I published them in this version). If you wish to alter
fixedcols/fixedrows in OnClick, just assign new values to SavedFixedCols/SavedFixedRows.
They set fixedcols/fixedrows when OnClick finishes}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, Clipbrd, mStrList;

type
  HandleFixed = (hfNot,hfSmart,hfYes);
  mSortState = (Unsorted,CompleteRows,CompleteCols,SingleCol,SingleRow,AllRows,AllCols);
  TMSortType = (soString,soStringCaseSensitive,soNumeric,soDate);
  TmSelectMethod = (smRange,smCompleteRows,smFree);
  TmMouseButton = (mpNone,mpLeft,mpElse);
  TonDrawCell = procedure(Sender:TObject; ACol,ARow:Longint; Rect:TRect; State:TGridDrawState{; var Text:string}) of object;
  TonFixedColClick = procedure(Sender:TObject; ACol:Longint) of object;
  TonFixedRowClick = procedure(Sender:TObject; ARow:Longint) of object;
  TmStrGrid = class(TStringgrid)
  private
    { Private-Deklarationen }
    fKeyType:TMSortType;
    fKeyPos:Integer;
    fKeyLen:Integer;
    fUseFixed:HandleFixed;
    SearchKey:String;
    FirstSearchRow,FirstSearchCol,ActSearchRow,ActSearchCol,LastSearchRow,LastSearchCol:Integer;
    SearchRowWise,SearchIsActive:Boolean;
    fSearchSubstring:Boolean;
    fEnableClipboardShortcuts:Boolean;
    fPasteToCursor:Boolean;
    fSortDescending:Boolean;
    fReverseSearch:Boolean;
    fSearchExists:Boolean;
    SortState:mSortState;
    SortArg:integer;
    fSortIndex:TmStrList;
    ColAnfang:integer;                     {Sort complete rows: Which col to begin}
    RowAnfang:integer;                     {Sort complete cols: Which row to begin}
    fDateTimeFormat:String;
    fShortYearExpand:Boolean;
    fShortYearBorder:Integer;
    {clicking}
    clicked:boolean;                       {While in Clicking on fixed cells, the fixed rows/cols ...}
    fSavedFixedRows:Integer;               {... must be set to zero. The mStrGrid methods called in OnClick ...}
    fSavedFixedCols:Integer;               {... have to remember the actual fixedrows/fixedcols}
    ClickedCol,ClickedRow:integer;
    ColumnSortState,
    RowSortState:AnsiString;
    fClickSortCols:Boolean;
    fClickSortRows:Boolean;
    ProblematicFixedClick:Boolean;
    FOnFixedColClick:TonFixedColClick;
    FOnFixedRowClick:TonFixedRowClick;
    {selecting}
    FOnDrawCell:TonDrawCell;
    fSelectMethod:TmSelectMethod;
    NewSelectionInfo:TStringList;
    FullUnSelectedRowInfo,FullSelectedRowInfo:AnsiString;
    LastRange:TGridRect;
    MousePressed:TmMouseButton;
    KeyPressed:Boolean;
    fSelectionColor:TColor;
    fSelectionTextColor:TColor;
    fSelectionFreezed:Boolean;
    function GetFixedRows:Integer;
    function GetFixedCols:Integer;
    procedure CalcDrawInfo(var DrawInfo: TGridDrawInfo);
    procedure CalcSizingState(X, Y: Integer; var State: TGridState;
      var Index: Longint; var SizingPos, SizingOfs: Integer;
      var FixedInfo: TGridDrawInfo);
    function CalcCoordFromPoint(X, Y: Integer;
      const DrawInfo: TGridDrawInfo): TGridCoord;
    procedure CalcDrawInfoXY(var DrawInfo: TGridDrawInfo;
      UseWidth, UseHeight: Integer);
    procedure CalcFixedInfo(var DrawInfo: TGridDrawInfo);
    function GetColWidths(Index: Longint): Integer;
    function GetRowHeights(Index: Longint): Integer;
    procedure SetReverseSearch(value:Boolean);
    procedure SetDateTimeFormat(value:String);
    function GetDateTimeFormat:String;
    function GetShortYearExpand:Boolean;
    procedure SetShortYearExpand(value:Boolean);
    function GetShortYearBorder:Integer;
    procedure SetShortYearBorder(value:Integer);
    {selecting}
    function  GetNewSelected(ACol, ARow: Integer):Boolean;
    procedure SetNewSelected(ACol, ARow: Integer; Const value:Boolean);
    procedure SetNewSelectMethod(value:TmSelectMethod);
    procedure CreateRowInformations;
    procedure UnSelect_NewSelections;
    procedure Copy_OldSelection_To_NewSelection;
    procedure Set_NewSelection_to_be_OldSelection;
    function  ChkGridSize:boolean;
    function  ChkSel_True:boolean;
    function  ChkSel_False:boolean;
    procedure UnSelectCompleteRow(r:integer);
    procedure SelectCompleteRow(r:integer);
    procedure FollowRange_CompleteRowSelect;
    procedure FollowRange_FreeSelect;
  protected
    { Protected-Deklarationen }
    procedure MouseDown(Button:TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure MouseUp(Button:TMouseButton; Shift:TShiftState; X,Y:Integer); override;
    procedure Click; override;
    procedure KeyPress(var Key:Char); override;
    procedure UndoCompRowSort;
    procedure UndoCompColSort;
    procedure UndoRowSort;
    procedure UndoColSort;
    procedure UndoAllRowsSort;
    procedure UndoAllColsSort;
    {selecting}
    procedure KeyDown(var Key:Word; Shift:TShiftState); override;
    procedure KeyUp(var Key:Word; Shift:TShiftState); override;
  public
    { Public-Deklarationen }
    constructor Create(AOwner: TComponent);override;
    procedure CutToClipboard;
    procedure CopyToClipboard;
    procedure PasteFromClipboard;
    destructor destroy;override;
    {clicking}
    procedure FixedColClick(ACol:Longint);
    procedure FixedRowClick(ARow:Longint);
    {selecting}
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState); override;
    property Selected[ACol,ARow:Integer]:Boolean read GetNewSelected write SetNewSelected;
    procedure UnSelect;
  published
    { Published-Deklarationen }
    procedure InsertCols(where,howmuch:Integer);
    procedure InsertRows(where,howmuch:Integer);
    procedure DeleteCols(where,howmuch:Integer);
    procedure DeleteRows(where,howmuch:Integer);
    procedure AddRow(contents:String; delimiter:Char);
    procedure AddCol(contents:String;  delimiter:Char);
    function ModifyRow(which:integer; contents,delimiter:string):integer;
    function ModifyCol(which:integer; contents,delimiter:string):integer;
    function FindFirst(Key:String; RowWise:Boolean; VAR ResultCol,ResultRow:Integer):Boolean;
    function FindFirstInRow(Key:String; searchRow:Integer; VAR ResultCol,ResultRow:Integer):Boolean;
    function FindFirstInCol(Key:String; searchCol:Integer; VAR ResultCol,ResultRow:Integer):Boolean;
    function FindNext(VAR ResultCol,ResultRow:Integer):Boolean;
    procedure SortCompleteColumns(KeyRow:integer);
    procedure SortCompleteRows(KeyCol:integer);
    procedure SortRow(ThisRow:integer);
    procedure SortCol(ThisCol:integer);
    procedure SortAllRows;
    procedure SortAllCols;
    procedure ReorderCompleteRows(KeyCol:integer; OrderValues:TStrings);
    procedure UndoSort;
    procedure LoadFromFile(FileName,ColSep:string);
    procedure AppendRowsFromFile(FileName,ColSep:String; ScipRows:Integer);
    procedure SaveToFile(FileName,ColSep:String);
    procedure assign(x:TStringGrid);
    property KeyType:TMSortType read fKeyType write fKeyType;
    property KeyPos:Integer read fKeyPos write fKeyPos;
    property KeyLen:Integer read fKeyLen write fKeyLen;
    property UseFixed:HandleFixed read fUseFixed write fUseFixed;
    property SearchSubstring:Boolean read fSearchSubstring write fSearchSubstring;
    property EnableClipboardShortcuts:Boolean read fEnableClipboardShortcuts write fEnableClipboardShortcuts;
    property OnKeyPress;
    property OnMouseDown;
    property OnMouseUp;
    property OnClick;
    property PasteToCursor:Boolean read fPasteToCursor write fPasteToCursor;
    property ClickSortCols:Boolean read fClickSortCols write fClickSortCols;
    property ClickSortRows:Boolean read fClickSortRows write fClickSortRows;
    property SortDescending:Boolean read fSortDescending write fSortDescending;
    property ReverseSearch:Boolean read fReverseSearch write SetReverseSearch;
    property SearchExists:Boolean read fSearchExists;
    property SortIndex:TmStrList read fSortIndex;
    property SavedFixedRows:Integer read GetFixedRows write fSavedFixedRows;
    property SavedFixedCols:Integer read GetFixedCols write fSavedFixedCols;
    property DateTimeFormat:String read GetDateTimeFormat write SetDateTimeFormat;
    property ShortYearExpand:Boolean read GetShortYearExpand write SetShortYearExpand;
    property ShortYearBorder:Integer read GetShortYearBorder write SetShortYearBorder;
    {clicking}
    property OnFixedColClick:TonFixedColClick read fonFixedColClick write fonFixedColClick;
    property OnFixedRowClick:TonFixedRowClick read fonFixedRowClick write fonFixedRowClick;
    {selecting}
    property SelectMethod:TmSelectMethod read fSelectMethod write SetNewSelectMethod;
    property OnDrawCell:TonDrawCell read fonDrawCell write fonDrawCell;
    property SelectionColor:TColor read fSelectionColor write fSelectionColor;
    property SelectionTextColor:TColor read fSelectionTextColor write fSelectionTextColor;
    property FreezeSelection:Boolean read fSelectionFreezed write fSelectionFreezed;
  end;

  mStrGridError = class(Exception);

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Additional', [TmStrGrid]);
end;

FUNCTION rVAL(CONST von:String):Real;
  {Reads a real value out of the string without an error message.
   Returns 0 if no numeric value}
  VAR bis,err:Integer; nach:Real;
  BEGIN VAL(von,nach,bis);
        IF bis>0 THEN VAL(copy(von,1,bis-1),nach,err);
        rVAL:=nach
  END;{rVAL}

function bool2char(b:boolean):char;
{Results in '0' or '1'}
begin result:=chr(48+ord(b)) end;

function char2bool(c:char):boolean;
begin result:=(c<>'0') end;

constructor TmStrGrid.Create(AOwner: TComponent);
var r:integer; RowInfo:AnsiString;
begin inherited Create(AOwner);
      fKeyType:=soString;
      fKeyPos:=1;
      fKeyLen:=MaxInt;
      clicked:=false;
      fSearchExists:=false;
      SortState:=Unsorted;
      fSortIndex:=TMStrList.create;
      fDateTimeFormat:=fSortIndex.DateTimeFormat;
      fShortYearExpand:=fSortIndex.ShortYearExpand;
      fShortYearBorder:=fSortIndex.ShortYearBorder;
      fSelectionColor:=clRed;
      fSelectionTextColor:=clWhite;
      {selection}
      NewSelectionInfo:=TStringList.create;
      MousePressed:=mpNone;
      KeyPressed:=false;
      {Fill in}
      CreateRowInformations;
      NewSelectionInfo.clear;
      for r:=0 to RowCount - 1 do NewSelectionInfo.Add(FullUnSelectedRowInfo);
      RowInfo:=NewSelectionInfo[FixedRows];
      RowInfo[succ(FixedCols)]:='1';
      NewSelectionInfo[FixedRows]:=RowInfo;
      LastRange:=Selection;
      ColumnSortState:='d';
      RowSortState:='d';
end;

destructor TmStrGrid.destroy;
begin fSortIndex.free;
      {selection}
      NewSelectionInfo.free;
      inherited
end;

function TmStrGrid.GetDateTimeFormat:String;
begin result:=fSortIndex.DateTimeFormat end;

procedure TmStrGrid.SetDateTimeFormat(value:String);
begin if value<>fDateTimeFormat then
       begin fDateTimeFormat:=value;
             fSortIndex.DateTimeFormat:=value
end    end;

function TmStrGrid.GetShortYearExpand:Boolean;
begin result:=fSortIndex.ShortYearExpand end;

procedure TmStrGrid.SetShortYearExpand(value:Boolean);
begin if value<>fShortYearExpand then
      begin fShortYearExpand:=value;
            fSortIndex.ShortYearExpand:=value
end   end;

function TmStrGrid.GetShortYearBorder:Integer;
begin result:=fSortIndex.ShortYearBorder end;

procedure TmStrGrid.SetShortYearBorder(value:Integer);
begin if value<>fShortYearBorder then
      begin fShortYearBorder:=value;
            fSortIndex.ShortYearBorder:=value
end   end;

procedure TmStrGrid.SortCompleteColumns(KeyRow:integer);
{Spalten sortieren nach der Zeile KeyRow}
var i,colanfang:integer;
begin case fUseFixed of
       hfNot  :begin rowanfang:=GetFixedRows; colanfang:=GetFixedCols end;
       hfSmart:begin rowanfang:=0;            colanfang:=GetFixedCols end;
       {hfYes} else  rowanfang:=0;            colanfang:=0
      end;
      fSortIndex.clear;
      fSortIndex.KeyPos:=fKeyPos;
      fSortIndex.KeyLen:=fKeyLen;
      fSortIndex.KeyType:=TLSortType(fKeyType);
      fSortIndex.ScipFirst:=colanfang;
      fSortIndex.SortDescending:=fSortDescending;
      {get key row}
      fSortIndex.addstrings(Rows[KeyRow]);
      {Sortierindizees bilden}
      for i:=0 to colcount-1 do
       fSortIndex.objects[i]:=TObject(i);
      {Liste sortieren}
      fSortIndex.sorted:=true;
      fSortIndex.sorted:=false;
      fSortIndex.RestoreOrderInGroups;
      {Indizees lesbar machen}
      with fSortIndex do
       for i:=0 to count-1 do
        strings[i]:=inttostr(integer(objects[i]));
      {Spalten vertauschen}
      UndoCompColSort;
      SortState:=CompleteCols;
end;

procedure TmStrGrid.UndoCompColSort;
var puffer,index:TStringList; i,j,k:integer;
begin index:=TStringList.create;
      index.addstrings(fSortIndex);
      {Resort the Sort Index, so that the index leads to unsort}
      with Index do
      for i:=0 to count-1 do
        begin fSortIndex.objects[integer(objects[i])]:=TObject(i);
              fSortIndex.strings[integer(objects[i])]:=inttostr(i);
        end;
      puffer:=TStringList.create;
      repeat {Suche einen Zyklus}
             i:=0;
             with Index do
             begin while (i<count) and (i=integer(objects[i])) do inc(i);
                   if i=count then break;
                   {Tausche Zyklus}
                   puffer.clear;
                   puffer.addstrings(cols[i]);
                   repeat j:=Integer(objects[i]);
                          objects[i]:=TObject(i);
                          if integer(objects[j])=j then break;
                          for k:=rowanfang to RowCount-1 do
                            cells[i,k]:=cells[j,k];
                          i:=j;
                   until false;
             end;
             for k:=rowanfang to RowCount-1 do
               begin cells[i,k]:=puffer.strings[k];
                     objects[i,k]:=puffer.objects[k]
               end;
      until false;
      puffer.free
end;

procedure TmStrGrid.SortCompleteRows(KeyCol:integer);
{Zeilen sortieren nach der Spalte KeyCol}
var i,rowanfang:integer;
begin case fUseFixed of
       hfNot  :begin rowanfang:=GetFixedRows; colanfang:=GetFixedCols end;
       hfSmart:begin rowanfang:=GetFixedRows; colanfang:=0 end;
       {hfYes} else  rowanfang:=0;            colanfang:=0
      end;
      fSortIndex.clear;
      fSortIndex.KeyPos:=fKeyPos;
      fSortIndex.KeyLen:=fKeyLen;
      fSortIndex.KeyType:=TLSortType(fKeyType);
      fSortIndex.ScipFirst:=rowanfang;
      fSortIndex.SortDescending:=fSortDescending;
      {get key col}
      fSortIndex.addstrings(cols[KeyCol]);
      {Sortierindizees bilden}
      for i:=0 to rowcount-1 do
       fSortIndex.objects[i]:=TObject(i);
      {Liste sortieren}
      fSortIndex.sorted:=true;
      fSortIndex.sorted:=false;
      fSortIndex.RestoreOrderInGroups;
      {Indizees lesbar machen}
      with fSortIndex do
       for i:=0 to count-1 do
        strings[i]:=inttostr(integer(objects[i]));
      {Zeilen vertauschen}
      UndoCompRowSort;
      SortState:=CompleteRows;
end;

procedure TmStrGrid.UndoCompRowSort;
var puffer,index:TStringList; i,j,k:integer;
begin index:=TStringList.create;
      index.addstrings(fSortIndex);
      {Resort the Sort Index, so that the index leads to unsort}
      with index do
      for i:=0 to count-1 do
        begin fSortIndex.objects[integer(objects[i])]:=TObject(i);
              fSortIndex.strings[integer(objects[i])]:=inttostr(i);
        end;
      puffer:=TStringList.create;
      repeat {Suche einen Zyklus}
             i:=0;
             with index do
             begin while (i<count) and (i=integer(objects[i])) do inc(i);
                   if i=count then break;
                   {Tausche Zyklus}
                   puffer.clear;
                   puffer.addstrings(rows[i]);
                   repeat j:=Integer(objects[i]);
                          objects[i]:=TObject(i);
                          if integer(objects[j])=j then break;
                          for k:=ColAnfang to ColCount-1 do
                            cells[k,i]:=cells[k,j];
                          i:=j;
                   until false;
             end;
             for k:=ColAnfang to ColCount-1 do
               begin cells[k,i]:=puffer.strings[k];
                     objects[k,i]:=puffer.objects[k]
               end;
      until false;
      puffer.free
end;

procedure TmStrGrid.ReorderCompleteRows(KeyCol:integer; OrderValues:TStrings);
{Zeilen umsortieren nach der Spalte KeyCol}
var i,p,rowanfang:integer;
begin case fUseFixed of
       hfNot  :begin rowanfang:=GetFixedRows; colanfang:=GetFixedCols end;
       hfSmart:begin rowanfang:=GetFixedRows; colanfang:=0 end;
       {hfYes} else  rowanfang:=0;            colanfang:=0
      end;
      if OrderValues.count<>RowCount-rowanfang then
        raise mStrGridError.Create('Incorrect list length of Order Values');
      with fSortIndex do
      begin clear;
            {get key col}
            addstrings(cols[KeyCol]);
            {Sortierindizees fuer fixed}
            for i:=0 to rowanfang-1 do
             objects[i]:=TObject(i);
            {Sortierindizees bilden durch Umordnung}
            for i:=rowanfang to RowCount-1 do
              begin p:=IndexOf(OrderValues.strings[i-rowanfang]);
                    if p=-1 then raise mStrGridError.Create('"'+OrderValues.strings[i-rowanfang]+'" not found in Grid');
                    objects[i]:=TObject(p);
      end     end;
      {Zeilen vertauschen}
      UndoCompRowSort;
      SortState:=CompleteRows;
end;

procedure TmStrGrid.SortRow(ThisRow:integer);
var i,colanfang:integer;
begin case fUseFixed of
       hfNot  :colanfang:=GetFixedCols;
       hfSmart:colanfang:=GetFixedCols; else
       {hfYes} colanfang:=0
      end;
      SortArg:=ThisRow;
      fSortIndex.clear;
      fSortIndex.KeyPos:=fKeyPos;
      fSortIndex.KeyLen:=fKeyLen;
      fSortIndex.KeyType:=TLSortType(fKeyType);
      fSortIndex.ScipFirst:=colanfang;
      fSortIndex.SortDescending:=fSortDescending;
      {Go!}
      fSortIndex.addstrings(Rows[ThisRow]);
      {Sortierindizees bilden}
      for i:=0 to colcount-1 do
       fSortIndex.objects[i]:=TObject(i);
      {Liste sortieren}
      fSortIndex.sorted:=true;
      fSortIndex.sorted:=false;
      {Indizees lesbar machen}
      with fSortIndex do
       for i:=0 to count-1 do
        strings[i]:=inttostr(integer(objects[i]));
      {Werte vertauschen}
      UndoRowSort;
      SortState:=SingleRow;
end;

procedure TmStrGrid.UndoRowSort;
var index:TStringList; i,j:integer; puffer:String; ptr:TObject;
begin index:=TStringList.create;
      index.addstrings(fSortIndex);
      {Resort the Sort Index, so that the index leads to unsort}
      with Index do
      for i:=0 to count-1 do
        begin fSortIndex.objects[integer(objects[i])]:=TObject(i);
              fSortIndex.strings[integer(objects[i])]:=inttostr(i);
        end;
      repeat {Suche einen Zyklus}
             i:=0;
             with Index do
             while (i<count) and (i=integer(objects[i])) do inc(i);
             if i=Index.count then break;
             {Tausche Zyklus}
             puffer:=cells[i,SortArg]; ptr:=objects[i,SortArg];
             with index do
             repeat j:=Integer(objects[i]);
                    objects[i]:=TObject(i);
                    if integer(objects[j])=j then break;
                    cells[i,SortArg]:=cells[j,SortArg];
                    i:=j;
             until false;
             cells[i,SortArg]:=puffer;
             objects[i,SortArg]:=ptr
      until false;
end;

procedure TmStrGrid.SortCol(ThisCol:integer);
var i,rowanfang:integer;
begin case fUseFixed of
       hfNot  :rowanfang:=GetFixedRows;
       hfSmart:rowanfang:=GetFixedRows; else
       {hfYes} rowanfang:=0;
      end;
      SortArg:=ThisCol;
      fSortIndex.clear;
      fSortIndex.KeyPos:=fKeyPos;
      fSortIndex.KeyLen:=fKeyLen;
      fSortIndex.KeyType:=TLSortType(fKeyType);
      fSortIndex.ScipFirst:=rowanfang;
      fSortIndex.SortDescending:=fSortDescending;
      {Go!}
      fSortIndex.addstrings(Cols[ThisCol]);
      {Sortierindizees bilden}
      for i:=0 to rowcount-1 do
       fSortIndex.objects[i]:=TObject(i);
      {Liste sortieren}
      fSortIndex.sorted:=true;
      fSortIndex.sorted:=false;
      {Indizees lesbar machen}
      with fSortIndex do
       for i:=0 to count-1 do
        strings[i]:=inttostr(integer(objects[i]));
      {Werte vertauschen}
      UndoColSort;
      SortState:=SingleCol;
end;

procedure TmStrGrid.UndoColSort;
var index:TStringList; i,j:integer; puffer:String; ptr:TObject;
begin index:=TStringList.create;
      index.addstrings(fSortIndex);
      {Resort the Sort Index, so that the index leads to unsort}
      with Index do
      for i:=0 to count-1 do
        begin fSortIndex.objects[integer(objects[i])]:=TObject(i);
              fSortIndex.strings[integer(objects[i])]:=inttostr(i);
        end;
      repeat {Suche einen Zyklus}
             i:=0;
             with Index do
             while (i<count) and (i=integer(objects[i])) do inc(i);
             if i=Index.count then break;
             {Tausche Zyklus}
             puffer:=cells[SortArg,i]; ptr:=objects[SortArg,i];
             with index do
             repeat j:=Integer(objects[i]);
                    objects[i]:=TObject(i);
                    if integer(objects[j])=j then break;
                    cells[SortArg,i]:=cells[SortArg,j];
                    i:=j;
             until false;
             cells[SortArg,i]:=puffer;
             objects[SortArg,i]:=ptr
      until false;
end;

procedure TmStrGrid.SortAllRows;
{Sorting all rows independent}
var i,r,colanfang:integer; index:TmStrList;
begin case fUseFixed of
       hfNot  :begin rowanfang:=GetFixedRows; colanfang:=GetFixedCols end;
       hfSmart:begin rowanfang:=0;            colanfang:=GetFixedCols end;
       {hfYes} else  rowanfang:=0;            colanfang:=0
      end;
      fSortIndex.clear;
      index:=TmStrList.create;
      with index do
      begin KeyPos:=fKeyPos;
            KeyLen:=fKeyLen;
            KeyType:=TLSortType(fKeyType);
            ScipFirst:=colanfang;
            SortDescending:=fSortDescending;
            for r:=Rowanfang to RowCount-1 do
            begin clear;
                  addstrings(Rows[r]);
                  {Sortierindizees bilden}
                  for i:=0 to colcount-1 do
                   objects[i]:=TObject(i);
                  {Liste sortieren}
                  sorted:=true;
                  sorted:=false;
                  {Sortierindizees speichern}
                  for i:=0 to count-1 do
                   fSortIndex.addobject(inttostr(integer(objects[i])),objects[i]);
            end;
      SortState:=AllRows;
      UndoAllRowsSort;
end   end;

procedure TmStrGrid.UndoAllRowsSort;
var fullindex,index:TStringList; i,j,aktrow,offset:integer; puffer:String; ptr:TObject;
begin fullindex:=TStringList.create;
      fullindex.addstrings(fSortIndex);
      index:=TStringList.create;
      offset:=0;
      for aktrow:=Rowanfang to RowCount-1 do
      begin with index do
            begin {get the indizees of the row i}
                  clear;
                  for i:=0 to Colcount-1 do
                    Index.addobject('',fullindex.objects[offset+i]);
                  {Resort the Sort Index, so that the index leads to unsort}
                  for i:=0 to count-1 do
                    begin fSortIndex.objects[offset+integer(objects[i])]:=TObject(i);
                          fSortIndex.strings[offset+integer(objects[i])]:=inttostr(i);
                    end;
                  offset:=offset+count;
            end;
            repeat {Suche einen Zyklus}
                   i:=0;
                   with Index do
                   while (i<count) and (i=integer(objects[i])) do inc(i);
                   if i=Index.count then break;
                   {Tausche Zyklus}
                   puffer:=cells[i,aktrow]; ptr:=objects[i,aktrow];
                   with index do
                   repeat j:=Integer(objects[i]);
                          objects[i]:=TObject(i);
                          if integer(objects[j])=j then break;
                          cells[i,aktrow]:=cells[j,aktrow];
                          i:=j;
                   until false;
                   cells[i,aktrow]:=puffer;
                   objects[i,aktrow]:=ptr
            until false;
end   end;

procedure TmStrGrid.SortAllCols;
{Sort all cols independend}
var i,c,rowanfang:integer; index:TmStrList;
begin case fUseFixed of
       hfNot  :begin rowanfang:=GetFixedRows; colanfang:=GetFixedCols end;
       hfSmart:begin rowanfang:=GetFixedRows; colanfang:=0 end;
       {hfYes} else  rowanfang:=0;            colanfang:=0
      end;
      fSortIndex.clear;
      index:=TmStrList.create;
      with index do
      begin KeyPos:=fKeyPos;
            KeyLen:=fKeyLen;
            KeyType:=TLSortType(fKeyType);
            ScipFirst:=rowanfang;
            SortDescending:=fSortDescending;
            for c:=colanfang to ColCount-1 do
            begin clear;
                  addstrings(Cols[c]);
                  {Sortierindizees bilden}
                  for i:=0 to Rowcount-1 do
                   objects[i]:=TObject(i);
                  {Liste sortieren}
                  sorted:=true;
                  sorted:=false;
                  {Sortierindizees speichern}
                  for i:=0 to count-1 do
                   fSortIndex.addobject(inttostr(integer(objects[i])),objects[i]);
            end;
      SortState:=AllCols;
      UndoAllColsSort;
end   end;

procedure TmStrGrid.UndoAllColsSort;
var fullindex,index:TStringList; i,j,aktcol,offset:integer; puffer:String; ptr:TObject;
begin fullindex:=TStringList.create;
      fullindex.addstrings(fSortIndex);
      index:=TStringList.create;
      offset:=0;
      for aktcol:=Colanfang to ColCount-1 do
      begin with index do
            begin {get the indizees of the col i}
                  clear;
                  for i:=0 to Rowcount-1 do
                    Index.addobject('',fullindex.objects[offset+i]);
                  {Resort the Sort Index, so that the index leads to unsort}
                  for i:=0 to count-1 do
                    begin fSortIndex.objects[offset+integer(objects[i])]:=TObject(i);
                          fSortIndex.strings[offset+integer(objects[i])]:=inttostr(i);
                    end;
                  offset:=offset+count;
            end;
            repeat {Suche einen Zyklus}
                   i:=0;
                   with Index do
                   while (i<count) and (i=integer(objects[i])) do inc(i);
                   if i=Index.count then break;
                   {Tausche Zyklus}
                   puffer:=cells[aktcol,i]; ptr:=objects[aktcol,i];
                   with index do
                   repeat j:=Integer(objects[i]);
                          objects[i]:=TObject(i);
                          if integer(objects[j])=j then break;
                          cells[aktcol,i]:=cells[aktcol,j];
                          i:=j;
                   until false;
                   cells[aktcol,i]:=puffer;
                   objects[aktcol,i]:=ptr
            until false;
end   end;

procedure TmStrGrid.UndoSort;
begin case SortState of
      CompleteRows:UndoCompRowSort;
      CompleteCols:UndoCompColSort;
      SingleCol:UndoColSort;
      SingleRow:UndoRowSort;
      AllRows:UndoAllRowsSort;
      AllCols:UndoAllColsSort;
end   end;

function TmStrGrid.ModifyRow(which:integer; contents,delimiter:string):integer;
var p,c,len:integer;
begin c:=0;
      len:=length(delimiter)-1;
      repeat p:=pos(delimiter,contents);
             if p=0 then break;
             cells[c,which]:=copy(contents,1,p-1);
             system.delete(contents,1,p+len);
             inc(c)
      until false;
      cells[c,which]:=contents;
      ModifyRow:=c+1
end;

function TmStrGrid.ModifyCol(which:integer; contents,delimiter:string):integer;
var p,r,len:integer;
begin r:=0;
      len:=length(delimiter)-1;
      repeat p:=pos(delimiter,contents);
             if p=0 then break;
             cells[which,r]:=copy(contents,1,p-1);
             system.delete(contents,1,p+len);
             inc(r)
      until false;
      cells[which,r]:=contents;
      ModifyCol:=r+1
end;

procedure TmStrGrid.LoadFromFile(FileName,ColSep:string);
var f:textfile; zeile:string; MaxCol,z,s:integer;
begin assignfile(f,FileName);
      {$I-}reset(f);{$I+}
      if IOResult <> 0 then
        raise mStrGridError.Create('File '+Filename+' not found');
      z:=-1;
      MaxCol:=0;
      while not eof(f) do
        begin readln(f,zeile);
              if zeile='' then continue;
              inc(z);
              if z>GetFixedRows then RowCount:=z+1;
              s:=ModifyRow(z,zeile,ColSep);
              if s>MaxCol then MaxCol:=s;
        end;
      RowCount:=z+1;
      ColCount:=MaxCol;
      closefile(f)
end;

procedure TmStrGrid.AppendRowsFromFile(FileName,ColSep:String; ScipRows:Integer);
var f:textfile; zeile:string; z:integer;
begin assignfile(f,FileName);
      {$I-}reset(f);{$I+}
      if IOResult <> 0 then
        raise mStrGridError.Create('File '+Filename+' not found');
      for z:=1 to ScipRows do
       if not eof(f) then
         readln(f,zeile);
      z:=RowCount-1;
      while not eof(f) do
        begin readln(f,zeile);
              if zeile='' then continue;
              inc(z);
              if z>GetFixedRows then RowCount:=z+1;
              ModifyRow(z,zeile,ColSep);
        end;
      RowCount:=z+1;
      closefile(f)
end;

procedure TmStrGrid.SaveToFile(FileName,ColSep:String);
var f:textfile; z,s:integer;
begin assignfile(f,FileName);
      rewrite(f);
      for z:=0 to RowCount-1 do
        begin for s:=0 to ColCount-1 do
                begin if s>0 then write(f,ColSep);
                      write(f,cells[s,z]);
                end;
              writeln(f)
        end;
      closefile(f)
end;

procedure TmStrGrid.CutToClipboard;
var r,c:integer; RowInf:AnsiString;
begin CopyToClipboard;
      for r:=0 to RowCount-1 do
       begin RowInf:=NewSelectionInfo[r];
             for c:=length(RowInf) downto 1 do
              if RowInf[c]='1' then cells[pred(c),r]:=''
end    end;

procedure TmStrGrid.CopyToClipboard;
var r,c,p:integer; RowInf,buffer:AnsiString; xmin,xmax,ymin,ymax:integer;
    puffer:PChar;
begin {Find rectangle borders for selected cells}
      xmin:=maxint; xmax:=-1; ymin:=maxint;ymax:=-1;
      for r:=0 to RowCount-1 do
      begin RowInf:=NewSelectionInfo[r];
            if RowInf<>FullUnSelectedRowInfo then
            begin if r<ymin then ymin:=r; if r>ymax then ymax:=r;
                  p:=pos('1',RowInf); if p=0 then raise mStrGridError.Create('Incorrect New Selection1');
                  if p<xmin then xmin:=p; if p>xmax then xmax:=p;
                  for c:=length(RowInf) downto p do
                    if RowInf[c]='1' then
                     begin if c>xmax then xmax:=c;
                           break
      end   end      end;
      {Fill rectangle now}
      buffer:='';
      for r:=ymin to ymax do
      begin RowInf:=NewSelectionInfo[r];
            for c:=xmin to pred(xmax) do
             if RowInf[c]='1'
              then buffer:=buffer+cells[pred(c),r]+#9
              else buffer:=buffer+#9;
            if RowInf[xmax]='1'
              then buffer:=buffer+cells[pred(xmax),r]+#13+#10
              else buffer:=buffer+#13+#10;
      end;
      {take last cr/lf away}
      setlength(buffer,length(buffer)-2);
      puffer:=#0;
      try GetMem(puffer,length(buffer));
          StrPCopy(puffer,buffer);
          ClipBoard.SetTextBuf(puffer);
          finally FreeMem(puffer,length(buffer));
      end;
end;

procedure TmStrGrid.PasteFromClipboard;
var Inhalt,Ende,Cr,Tab:PChar; rowinf:AnsiString;
    Eintrag:Array[0..255] of char;
    s,z,r,leftcol,toprow:integer;
    Data:THandle;
begin if not Clipboard.HasFormat(CF_TEXT) then exit;
      Clipboard.Open;
      Data:=GetClipboardData(CF_TEXT);
      try if Data<>0
           then inhalt:=PChar(GlobalLock(Data))
           else inhalt:=nil
      finally
           if Data<>0 then GlobalUnlock(Data);
           ClipBoard.Close;
      end;
      leftcol:=0; toprow:=0; {fake warnings}
      if inhalt=nil then exit;
      if fPasteToCursor
      then begin toprow:=row;
                 leftcol:=col;
           end
      else for r:=0 to RowCount-1 do
           begin RowInf:=NewSelectionInfo[r];
                 if RowInf<>FullUnSelectedRowInfo then
                 begin toprow:=r;
                       leftcol:=pos('1',RowInf)-1; if leftcol=-1 then raise mStrGridError.Create('Incorrect New Selection2');
                       break
           end   end;
      Ende:=StrScan(Inhalt,#0);
      s:=leftcol; z:=toprow;
      repeat {Zeile holen}
             Cr:=StrScan(inhalt,#13);
             if Cr=nil then Cr:=Ende;
             repeat Tab:=StrScan(inhalt,#9);
                    if (Tab>Cr) or (Tab=nil) then Tab:=Cr;
                    StrLCopy(Eintrag,Inhalt,Tab-Inhalt);
                    cells[s,z]:=StrPas(Eintrag);
                    inc(s);
                    if s=colCount then ColCount:=s;
                    Inhalt:=Tab+1
             until Tab=Cr;
             inc(Inhalt); {lf weg}
             s:=leftcol;
             inc(z);
             if z=RowCount then RowCount:=z+1
      until (Cr=Ende) or (inhalt^=#0)
end;

procedure TmStrGrid.InsertCols(where,howmuch:Integer);
var i:Integer;
begin ColCount:=ColCount+howmuch;
      for i:=ColCount-1 downto where+howmuch do
        Cols[i]:=Cols[i-howmuch];
      cols[where].clear;
      for i:=where+1 to where+howmuch-1 do
        Cols[i]:=Cols[where];
end;

procedure TmStrGrid.InsertRows(where,howmuch:Integer);
var i:Integer;
begin RowCount:=RowCount+howmuch;
      for i:=RowCount-1 downto where+howmuch do
        Rows[i]:=Rows[i-howmuch];
      Rows[where].clear;
      for i:=where+1 to where+howmuch-1 do
        Rows[i]:=Rows[where];
end;

procedure TmStrGrid.DeleteCols(where,howmuch:Integer);
var i:Integer;
begin if ColCount-where<howmuch then howmuch:=ColCount-where;
      for i:=where to ColCount-1 do
        Cols[i]:=Cols[i+howmuch];
      ColCount:=ColCount-howmuch;
end;

procedure TmStrGrid.DeleteRows(where,howmuch:Integer);
var i:Integer;
begin if RowCount-where<howmuch then howmuch:=RowCount-where;
      for i:=where to RowCount-1 do
        Rows[i]:=Rows[i+howmuch];
      RowCount:=RowCount-howmuch;
end;

procedure TmStrGrid.AddRow(contents:String; delimiter:Char);
begin RowCount:=RowCount+1;
      ModifyRow(RowCount-1,contents,delimiter);
end;

procedure TmStrGrid.AddCol(contents:String;  delimiter:Char);
begin ColCount:=ColCount+1;
      ModifyCol(ColCount-1,contents,delimiter);
end;

function TmStrGrid.FindFirst(Key:String; RowWise:Boolean; VAR ResultCol,ResultRow:Integer):Boolean;
begin SearchKey:=Key; SearchRowWise:=RowWise; SearchIsActive:=true;
      case fUseFixed of
       hfNot:begin FirstSearchRow:=GetFixedRows; FirstSearchCol:=GetFixedCols end;
       hfSmart:if rowwise
               then begin FirstSearchRow:=GetFixedRows; FirstSearchCol:=0; end
               else begin FirstSearchRow:=0; FirstSearchCol:=GetFixedCols; end;
       hfYes:begin FirstSearchRow:=0; FirstSearchCol:=0 end;
      end;
      LastSearchRow:=RowCount-1;
      LastSearchCol:=ColCount-1;
      if fReverseSearch
      then begin ActSearchRow:=LastSearchRow;
                 ActSearchCol:=LastSearchCol
           end
      else begin ActSearchRow:=FirstSearchRow;
                 ActSearchCol:=FirstSearchCol
           end;
      fSearchExists:=true;
      FindFirst:=FindNext(ResultCol,ResultRow);
end;

function TmStrGrid.FindFirstInRow(Key:String; searchRow:Integer; VAR ResultCol,ResultRow:Integer):Boolean;
begin SearchKey:=Key; SearchRowWise:=true; SearchIsActive:=true;
      FirstSearchRow:=searchRow; ActSearchRow:=searchRow; LastSearchRow:=searchRow;
      if fUseFixed=hfNot
      then FirstSearchCol:=GetFixedCols
      else FirstSearchCol:=0;
      if fReverseSearch
      then ActSearchCol:=LastSearchCol
      else ActSearchCol:=FirstSearchCol;
      LastSearchCol:=ColCount-1;
      fSearchExists:=true;
      FindFirstInRow:=FindNext(ResultCol,ResultRow)
end;

function TmStrGrid.FindFirstInCol(Key:String; searchCol:Integer; VAR ResultCol,ResultRow:Integer):Boolean;
begin SearchKey:=Key; SearchRowWise:=false; SearchIsActive:=true;
      FirstSearchCol:=searchCol; ActSearchCol:=searchCol; LastSearchCol:=searchCol;
      if fUseFixed=hfNot
      then FirstSearchRow:=GetFixedRows
      else FirstSearchRow:=0;
      if fReverseSearch
      then ActSearchRow:=LastSearchRow
      else ActSearchRow:=FirstSearchRow;
      LastSearchRow:=RowCount-1;
      fSearchExists:=true;
      FindFirstInCol:=FindNext(ResultCol,ResultRow);
end;

function TmStrGrid.FindNext(VAR ResultCol,ResultRow:Integer):Boolean;
var SearchReal:Real; SearchDate:TDateTime; found:boolean;
  function next:boolean;
  begin next:=false;
        if fReverseSearch
        then if SearchRowWise
             then begin dec(ActSearchCol);
                        if ActSearchCol<FirstSearchCol then
                           begin ActSearchCol:=LastSearchCol;
                                 dec(ActSearchRow);
                                 if ActSearchRow<FirstSearchRow then exit
                  end      end
             else begin dec(ActSearchRow);
                        if ActSearchRow<FirstSearchRow then
                           begin ActSearchRow:=LastSearchRow;
                                 dec(ActSearchCol);
                                 if ActSearchCol<FirstSearchCol then exit
                  end      end
        else if SearchRowWise
             then begin inc(ActSearchCol);
                        if ActSearchCol>LastSearchCol then
                           begin ActSearchCol:=FirstSearchCol;
                                 inc(ActSearchRow);
                                 if ActSearchRow>LastSearchRow then exit
                  end      end
             else begin inc(ActSearchRow);
                        if ActSearchRow>LastSearchRow then
                           begin ActSearchRow:=FirstSearchRow;
                                 inc(ActSearchCol);
                                 if ActSearchCol>LastSearchCol then exit
                  end      end;
        next:=true
  end;
begin if not SearchIsActive then
        begin FindNext:=false;
              ResultRow:=-1; ResultCol:=-1;
              exit
        end;
      SearchReal:=0; SearchDate:=0;
     {Suchwort konvertieren}
      case fKeyType of
       soNumeric:SearchReal:=Rval(SearchKey);
       soString:if fSearchSubstring then SearchKey:=UpperCase(SearchKey);
       soDate:SearchDate:=StrToDateTime(SearchKey);
      end;
      found:=false;
      if fSearchSubstring then
      repeat {Suche Passenden}
             case fKeyType of
              soString:if pos(SearchKey,UpperCase(cells[ActSearchCol,ActSearchRow]))>0 then
                        begin found:=true; break end;
              soStringCaseSensitive:if pos(SearchKey,cells[ActSearchCol,ActSearchRow])>0 then
                        begin found:=true; break end;
              soNumeric:if Rval(copy(cells[ActSearchCol,ActSearchRow],fKeyPos,fKeyLen))=SearchReal then
                        begin found:=true; break end;
              soDate:if StrToDateTime(copy(cells[ActSearchCol,ActSearchRow],fKeyPos,fKeyLen))=SearchDate then
                        begin found:=true; break end;
             end;
      until found or not next
      else
      repeat {Suche Passenden}
             case fKeyType of
              soString:if AnsiCompareText(copy(cells[ActSearchCol,ActSearchRow],fKeyPos,fKeyLen),SearchKey)=0 then
                        begin found:=true; break end;
              soStringCaseSensitive:if copy(cells[ActSearchCol,ActSearchRow],fKeyPos,fKeyLen)=SearchKey then
                        begin found:=true; break end;
              soNumeric:if Rval(copy(cells[ActSearchCol,ActSearchRow],fKeyPos,fKeyLen))=SearchReal then
                        begin found:=true; break end;
              soDate:if StrToDateTime(copy(cells[ActSearchCol,ActSearchRow],fKeyPos,fKeyLen))=SearchDate then
                        begin found:=true; break end;
             end;
      until found or not next;
      if found
       then begin ResultRow:=ActSearchRow; ResultCol:=ActSearchCol;
                  SearchIsActive:=next
            end
       else begin SearchIsActive:=false;
                  ResultRow:=-1; ResultCol:=-1
            end;
      FindNext:=found;
end;

procedure TmStrGrid.SetReverseSearch(value:Boolean);
var dummycol,dummyrow:integer;
begin if value<>fReverseSearch then
      begin fReverseSearch:=value;
            if SearchIsActive
            then findnext(dummycol,dummyrow)
            else if fSearchExists then
                   begin SearchIsActive:=true;
                         if fReverseSearch
                         then begin ActSearchRow:=LastSearchRow;
                                    ActSearchCol:=LastSearchCol
                              end
                         else begin ActSearchRow:=FirstSearchRow;
                                    ActSearchCol:=FirstSearchCol
end   end          end        end;

procedure TmStrGrid.KeyPress(var Key: Char);
begin if NOT EditorMode and EnableClipboardShortcuts then
      case key of
       ^c:CopyToClipboard;
       ^v:PasteFromClipboard;
       ^x:CutToClipboard;
      end;
      inherited KeyPress(key);
end;

{All following procedures/functions implement a new click behaviour}

procedure TmStrGrid.SetNewSelectMethod(value:TmSelectMethod);
var r:integer;
begin if value<>FSelectMethod then
      begin {Clear(create) New Selection}
            CreateRowInformations;
            NewSelectionInfo.clear;
            for r:=0 to RowCount - 1 do NewSelectionInfo.Add(FullUnSelectedRowInfo);
            FSelectMethod:=value;
            case FSelectMethod of
              smRange,smFree:Copy_OldSelection_To_NewSelection;
              smCompleteRows:with Selection do for r:=top to bottom do SelectCompleteRow(r);
            end;
            invalidate
end   end;

function TmStrGrid.GetFixedRows:Integer;
{This is to get the correct Fixedcols/rows value if the user click procedure calls mStrGrid methods}
begin if clicked then result:=fSavedFixedRows
                 else result:=FixedRows
end;

function TmStrGrid.GetFixedCols:Integer;
{This is to get the correct Fixedcols/rows value if the user click procedure calls mStrGrid methods}
begin if clicked then result:=fSavedFixedCols
                 else result:=FixedCols
end;

procedure TmStrGrid.Set_NewSelection_to_be_OldSelection;
var c,r:integer; RowInfo:AnsiString;
begin with selection do
      for r:=Top to Bottom do
      begin RowInfo:=NewSelectionInfo[r];
            for c:=Left to Right do RowInfo[succ(c)]:='1';
            NewSelectionInfo[r]:=RowInfo;
end   end;

procedure TmStrGrid.Copy_OldSelection_To_NewSelection;
var c,r:integer; RowInfo:AnsiString;
begin canvas.brush.color:=ClHighlight;
      canvas.font.color:=ClHighlighttext;
      with selection do
      for r:=Top to Bottom do
      begin RowInfo:=NewSelectionInfo[r];
            for c:=Left to Right do
              begin RowInfo[succ(c)]:='1';
                    inherited DrawCell(c,r,CellRect(c,r),[]);
              end;
            NewSelectionInfo[r]:=RowInfo;
end   end;

procedure TmStrGrid.CreateRowInformations;
begin setlength(FullUnSelectedRowInfo,ColCount);
      fillchar(pointer(LongInt(@(FullUnSelectedRowInfo[1])))^,ColCount,'0');
      setlength(FullSelectedRowInfo,ColCount);
      fillchar(pointer(LongInt(@(FullSelectedRowInfo[1])))^,ColCount,'1');
end;

procedure TmStrGrid.UnSelect_NewSelections;
var r,c_:integer; RowInfo:AnsiString; rect:TRect;
begin CreateRowInformations;
      {Find rows with selected cells and UnSelect_NewSelections these}
      for r:=0 to RowCount - 1 do
       if NewSelectionInfo[r]<>FullUnSelectedRowInfo then
        begin RowInfo:=NewSelectionInfo[r];
              canvas.brush.color:=ClBtnFace;
              canvas.font.color:=ClBtnText;
              for c_:=1 to FixedCols do
               if RowInfo[c_]='1' then
                begin rect:=CellRect(pred(c_),r);
                      inc(rect.top);
                      dec(rect.bottom);
                      inc(rect.left);
                      dec(rect.right);
                      inherited DrawCell(pred(c_),r, rect,[]);
                end;
              canvas.brush.color:=ClWindow;
              canvas.font.color:=ClWindowtext;
              for c_:=FixedCols+1 to ColCount do
               if RowInfo[c_]='1' then
                inherited DrawCell(pred(c_),r, CellRect(pred(c_),r),[]{AState});
        end;
      {Fill in}
      NewSelectionInfo.clear;
      for r:=0 to RowCount - 1 do NewSelectionInfo.Add(FullUnSelectedRowInfo);
end;

procedure TmStrGrid.UnSelect;
begin UnSelect_NewSelections;
      with selection do
      case FSelectMethod of
       smRange:Copy_OldSelection_To_NewSelection;
       smFree:begin Copy_OldSelection_To_NewSelection;
                    invalidate
              end;
       smCompleteRows:begin SelectCompleteRow(top);
                            invalidate
      end             end;
end;

function TmStrGrid.GetNewSelected(ACol, ARow: Integer):Boolean;
{Read function for SELECTED property}
begin ChkSel_True;
      if (ACol>=ColCount) or (ARow>=RowCount)
      then result:=false
      else result:=NewSelectionInfo[ARow][succ(ACol)]<>'0'
end;

procedure TmStrGrid.SetNewSelected(ACol, ARow: Integer; Const value:Boolean);
{Write procedure for SELECTED property}
var RowInf:AnsiString;
begin ChkSel_True;
      if (ACol>=ColCount) or (ARow>=RowCount) then exit;
      if NewSelectionInfo[ARow][succ(ACol)]<>bool2char(Value) then
       if fSelectMethod<>smCompleteRows
        then begin RowInf:=NewSelectionInfo[ARow];
                   RowInf[succ(ACol)]:=bool2char(Value);
                   NewSelectionInfo[ARow]:=RowInf;
                   if value then DrawCell(ACol, ARow, CellRect(ACol, ARow),[gdselected]{AState})
                            else DrawCell(ACol, ARow, CellRect(ACol, ARow),[]{AState});
             end
        else if value then SelectCompleteRow(ARow)
                      else UnSelectCompleteRow(ARow)
end;

function TmStrGrid.ChkGridSize:boolean;
var len:integer;
begin with NewSelectionInfo do
begin result:=(Count=RowCount);
      if result then
         begin len:=length(strings[0]);
               result:=(len=ColCount)
end end  end;

function TmStrGrid.ChkSel_True:boolean;
{Updates NewSelectionInfo to actual Grid size. Returns always True}
var c,r,len:integer; RowInfo:AnsiString;
begin with NewSelectionInfo do
begin result:=true;
      if Count<RowCount then
         begin {Additional Rows appeared}
               for r:=count to RowCount - 1 do Add(FullUnSelectedRowInfo);
         end else
      if Count>RowCount then
         begin {Some Rows disappeared}
               c:=count;
               for r:=count to RowCount - 1 do delete(c)
         end;
      if count=0 then exit;
      len:=length(strings[0]);
      if len<>ColCount then
         begin {Cols appeared or disappeared}
               CreateRowInformations;
               for r:=0 to Count - 1 do
                begin RowInfo:=strings[r];
                      len:=length(RowInfo);
                      if len>ColCount
                       then strings[r]:=copy(RowInfo,1,ColCount)
                       else begin for c:=len+1 to ColCount do
                                    RowInfo:=RowInfo+'0';
                                  strings[r]:=RowInfo
end end  end    end         end;

function TmStrGrid.ChkSel_False:boolean;
{Updates NewSelectionInfo to actual Grid size. Returns always False}
begin ChkSel_True;
      result:=false
end;

procedure TmStrGrid.UnSelectCompleteRow(r:integer);
var c:integer; rect:TRect;
begin {Draw fixed cells}
      canvas.brush.color:=ClBtnFace;
      canvas.font.color:=ClBtnText;
      for c:=0 to FixedCols-1 do
       begin rect:=CellRect(c,r);
             inc(rect.top);
             dec(rect.bottom);
             inc(rect.left);
             dec(rect.right);
             inherited DrawCell(c,r,rect,[]);
       end;
      {Draw unfixed cells}
      canvas.brush.color:=ClWindow;
      canvas.font.color:=ClWindowtext;
      for c:=FixedCols to Colcount-1 do
       inherited DrawCell(c,r,CellRect(c,r),[]);
      NewSelectionInfo[r]:=FullUnSelectedRowInfo;
end;

procedure TmStrGrid.SelectCompleteRow(r:integer);
var c:integer; rect:TRect;
begin canvas.brush.color:=fSelectionColor;
      canvas.font.color:=fSelectionTextColor;
      {Draw fixed cells}
      for c:=0 to FixedCols-1 do
       begin rect:=CellRect(c,r);
             inc(rect.top);
             dec(rect.bottom);
             inc(rect.left);
             dec(rect.right);
             inherited DrawCell(c,r,rect,[]);
       end;
      {Draw unfixed cells}
      for c:=FixedCols to Colcount-1 do
       inherited DrawCell(c,r,CellRect(c,r),[]);
      NewSelectionInfo[r]:=FullSelectedRowInfo;
end;

procedure TmStrGrid.FollowRange_CompleteRowSelect;
{Completed Rows are updated after the actual originalSelectRange
 LastRange must be valid}
var i:integer;
begin for i:=Selection.top to LastRange.top-1 do {new rows on top}
        SelectCompleteRow(i);
      for i:=LastRange.top to Selection.top-1 do {rows deselected on top}
        UnSelectCompleteRow(i);
      for i:=LastRange.bottom+1 to Selection.bottom do {new rows on bottom}
        SelectCompleteRow(i);
      for i:=Selection.bottom+1 to LastRange.bottom do {rows deselected on bottom}
        UnSelectCompleteRow(i);
      LastRange:=Selection;
end;

procedure TmStrGrid.FollowRange_FreeSelect;
{Cells are updated after the actual originalSelectRange
 LastRange must be valid}
var r,c:integer;
begin {Update Rows}
      for r:=Selection.top to LastRange.top-1 do {new rows on top}
        for c:=Selection.left to Selection.Right do SetNewSelected(c,r,true);
      for r:=LastRange.top to Selection.top-1 do {rows deselected on top}
        for c:=LastRange.left to LastRange.Right do SetNewSelected(c,r,false);
      for r:=LastRange.bottom+1 to Selection.bottom do {new rows on bottom}
        for c:=Selection.left to Selection.Right do SetNewSelected(c,r,true);
      for r:=Selection.bottom+1 to LastRange.bottom do {rows deselected on bottom}
        for c:=LastRange.left to LastRange.Right do SetNewSelected(c,r,false);
      {Update Columns}
      for c:=Selection.Left to LastRange.Left-1 do {new rows on top}
        for r:=Selection.Top to Selection.Bottom do SetNewSelected(c,r,true);
      for c:=LastRange.Left to Selection.Left-1 do {rows deselected on top}
        for r:=LastRange.Top to LastRange.Bottom do SetNewSelected(c,r,false);
      for c:=LastRange.Right+1 to Selection.Right do {new rows on bottom}
        for r:=Selection.Top to Selection.Bottom do SetNewSelected(c,r,true);
      for c:=Selection.Right+1 to LastRange.Right do {rows deselected on bottom}
        for r:=LastRange.Top to LastRange.Bottom do SetNewSelected(c,r,false);
      LastRange:=Selection;
end;

procedure TmStrGrid.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var Arow,Acol:LongInt; c,r:integer;
begin if clicked and (fClickSortCols or fClickSortRows) and ProblematicFixedClick then
        begin Click;
              clicked:=false;
              inherited;
              exit
        end;
      {This is to get the correct Fixedcols/rows value if the user click procedure calls xGrid methods}
      clicked:=false;
      ChkSel_True;
      MouseToCell(x,y,Acol,Arow);
      if Arow<0 then Arow:=RowCount-1;
      if (MousePressed=mpLeft) and not fSelectionFreezed then
      case FSelectMethod of
      smRange:begin {Just follow the official selection}
                    UnSelect_NewSelections;
                    Copy_OldSelection_To_NewSelection;
              end;
      smCompleteRows:{Just follow the official selection}
             if (Shift=[ssShift]) then
                FollowRange_CompleteRowSelect else
             if (Shift=[ssCtrl]) then
                begin {Flip rows}
                      for r:=Selection.top to Selection.bottom do
                       if GetNewSelected(col,r)
                        then UnSelectCompleteRow(r)
                        else SelectCompleteRow(r);
                      LastRange:=Selection;
                end
             else {Shift=[]}
                  begin UnSelect_NewSelections;
                        for r:=Selection.top to Selection.bottom do
                          SelectCompleteRow(r);
                        LastRange:=Selection;
                  end;
      smFree:if (Shift=[ssShift]) then
                FollowRange_FreeSelect else
             if (Shift=[ssCtrl]) then
                begin {Flip cells}
                      for r:=Selection.top to Selection.bottom do
                       for c:=Selection.left to Selection.right do
                        SetNewSelected(c,r,not GetNewSelected(c,r));
                      LastRange:=Selection;
                end else
             {if Shift=[] then}
                 begin UnSelect_NewSelections;
                       for r:=Selection.top to Selection.bottom do
                        for c:=Selection.left to Selection.right do
                         SetNewSelected(c,r,true);
                       LastRange:=Selection;
                 end;
      end;
      MousePressed:=mpNone;
      inherited;
end;{TmStrGrid.MouseUp}

procedure TmStrGrid.KeyDown(var Key:Word; Shift:TShiftState);
const CursorMove=[VK_HOME,VK_END,VK_PRIOR,VK_NEXT,VK_UP,VK_DOWN,VK_LEFT,VK_RIGHT];
begin KeyPressed:=true;
      inherited;
      if (Key in CursorMove) and not fSelectionFreezed then
      case FSelectMethod of
       smRange:begin {Just follow the official selection}
                     UnSelect_NewSelections;
                     Copy_OldSelection_To_NewSelection;
               end;
       smCompleteRows:FollowRange_CompleteRowSelect;
       smFree:FollowRange_FreeSelect;
      end;
end;{TmStrGrid.KeyDown}

procedure TmStrGrid.KeyUp(var Key:Word; Shift:TShiftState);
// []: selects only actual, [shift] selects range
// CompleteRows: Original range is invalid! (else the cursor is set to the end)
const CursorMove=[VK_HOME,VK_END,VK_PRIOR,VK_NEXT,VK_UP,VK_DOWN,VK_LEFT,VK_RIGHT];
begin ChkSel_True;
      KeyPressed:=false;
      if not fSelectionFreezed then
      case FSelectMethod of
        smRange:;{Action is in KeyDown}
        smCompleteRows:if Key in CursorMove then FollowRange_CompleteRowSelect;
        smFree:if Key in CursorMove then FollowRange_FreeSelect;
      end;
      inherited;
end;{TmStrGrid.KeyUp}

{This is the only modified procedure from TCustomGrid}

procedure TmStrGrid.MouseDown(Button:TMouseButton; Shift:TShiftState; X,Y:Integer);
var
  CellHit: TGridCoord;
  DrawInfo: TGridDrawInfo;
  DummySizingIndex:LongInt;
  DummySizingPos,DummySizingOfs:integer;
  OldOptions:TGridOptions;
begin
  if not (csDesigning in ComponentState) and CanFocus then
  begin
    SetFocus;
    if ValidParentForm(Self).ActiveControl <> Self then
    begin
      inherited;
      Exit;
    end;
  end;
  if (Button = mbRight) or (Button = mbMiddle) then
    MousePressed:=mpElse else
  if (Button = mbLeft) and (ssDouble in Shift) then
    DblClick
  else if Button = mbLeft then
  begin
    CalcDrawInfo(DrawInfo);
    { Check grid sizing }
    CalcSizingState(X, Y, FGridState, DummySizingIndex, DummySizingPos, DummySizingOfs, DrawInfo);
    if FGridState <> gsNormal then
    begin
      inherited;
      Exit;
    end;
    CellHit := CalcCoordFromPoint(X, Y, DrawInfo);
    if not fSelectionFreezed then
    case FSelectMethod of {We are here in MouseDown}
        smRange:if not (ssShift in Shift) then UnSelect_NewSelections;
        smCompleteRows:FollowRange_CompleteRowSelect;
        smFree:begin if ssCtrl in Shift
                     then FollowRange_FreeSelect
                     else begin UnSelect_NewSelections;
                                Copy_OldSelection_To_NewSelection
    end        end        end;
    MousePressed:=mpLeft;
    with CellHit do
    if (fClickSortRows and (X < FixedCols) and (X >=0)) or
       (fClickSortCols and (Y < FixedRows) and (Y >=0))
    then begin {Alter Fixed Col/Row Barrier}
               {This is to get the correct Fixedcols/rows value
                if the user click procedure calls mStrGrid methods}
               fSavedFixedRows:=FixedRows;
               fSavedFixedCols:=FixedCols;
               oldoptions:=options;
               options:=options-[goRowSelect];
               ProblematicFixedClick:=((FixedCols=0) and (X=0)) or
                                      ((FixedRows=0) and (Y=0));
               FixedRows:=0;
               FixedCols:=0;
               ClickedCol:=X;
               ClickedRow:=Y;
               clicked:=true;
               {The following two assignments call in the three stages inherited CustomGrid
                the methods SetCol which calls FocusCell which calls MoveCurrent.
                As it is not allowed to set COL or ROW into the fixed cells,
                the assignment is not done, if Y <=FixedRows (the same with X.)}
               row:=Y;
               col:=X;
               FixedRows:=fSavedFixedRows;
               FixedCols:=fSavedFixedCols;
               options:=oldoptions;
         end
    else inherited
  end
end;{TmStrGrid.MouseDown}

procedure TmStrGrid.Click;
{Branch here to normal or fixed click event}
begin if ProblematicFixedClick =  ((Col=ClickedCol) and (Row=ClickedRow)) then exit;
      if (row<SavedFixedRows) and (col<SavedFixedCols) then
        {clicked upper left, with existant fixed row and column} else
      if (row<SavedFixedRows) or (ProblematicFixedClick and (SavedFixedRows>0))
      then begin {Fixed Col was clicked}
                 if assigned(OnFixedColClick) then fOnFixedColClick(nil,col);
                 if fClickSortCols then FixedColClick(col);
           end else
      if (col<SavedFixedCols) or (ProblematicFixedClick and (SavedFixedCols>0))
      then begin {fixed row was clicked}
                 if assigned(OnFixedRowClick) then fOnFixedRowClick(nil,row);
                 if fClickSortRows then FixedRowClick(row)
           end
      else {Click in a non fixed cell}
           inherited
end;{TmStrGrid.Click}

procedure TmStrGrid.FixedColClick(ACol:Longint);
var oldlen:integer;
begin {Correct length of ColumnSortState}
      oldlen:=length(ColumnSortState);
      if ColCount>oldlen then
        begin {New Columns were added}
              SetLength(ColumnSortState,ColCount);
              fillchar(pointer(LongInt(@(ColumnSortState[oldlen+1])))^,ColCount-oldlen-1,'d');
        end else
      if ColCount<oldlen then
        {Columns were deleted}
        SetLength(ColumnSortState,ColCount);
      {Now sort it}
      if ColumnSortState[succ(Acol)]='a'
       then begin {ascending -> descending}
                  SortDescending:=true;
                  SortCompleteRows(Acol);
                  ColumnSortState[succ(Acol)]:='d'
            end
       else begin { -> ascending}
                  SortDescending:=false;
                  SortCompleteRows(Acol);
                  ColumnSortState[succ(Acol)]:='a'
end         end;

procedure TmStrGrid.FixedRowClick(ARow:Longint);
var oldlen:integer;
begin {Correct length of RowSortState}
      oldlen:=length(RowSortState);
      if RowCount>oldlen then
        begin {New Rows were added}
              SetLength(RowSortState,RowCount);
              fillchar(pointer(LongInt(@(RowSortState[oldlen+1])))^,RowCount-oldlen-1,'d');
        end else
      if RowCount<oldlen then
        {Rows were deleted}
        SetLength(RowSortState,RowCount);
      {Now sort it}
      if RowSortState[succ(Arow)]='a'
      then begin {ascending -> descending}
                 SortDescending:=true;
                 SortCompleteColumns(Arow);
                 RowSortState[succ(Arow)]:='d'
           end
      else begin { -> ascending}
                 SortDescending:=false;
                 SortCompleteColumns(Arow);
                 RowSortState[succ(Arow)]:='a'
end        end;

procedure TmStrGrid.DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
var olddefault:boolean; inhalt:string;
begin {If any program changes grid size, update New Selection size info}
      if (MousePressed<>mpLeft) and (not clicked) and (not KeyPressed) then
        ChkSel_True;
      inhalt:=Cells[ACol,ARow];
      olddefault:=defaultdrawing;
      {if assigned(FOnDrawCell) then FonDrawCell(self,Acol,Arow,Arect,Astate,inhalt);}
      if DefaultDrawing then
       if // Not at Design Time!
         not (csDesigning in ComponentState) and
         (fSelectMethod<>smRange) and
         ChkSel_True and
         char2bool(NewSelectionInfo[ARow][succ(Acol)])
         then begin {Show new cell selection}
                    canvas.brush.color:=fSelectionColor;
                    canvas.font.color:=fSelectionTextColor;
                    inherited DrawCell(ACol, ARow, ARect, AState);
              end else
       if (gdselected in Astate) and not fSelectionFreezed
         then begin {Show old cell selection}
                    canvas.brush.color:=ClHighlight;
                    canvas.font.color:=ClHighlighttext;
                    inherited DrawCell(ACol, ARow, ARect, AState);
              end else
       if gdfixed in Astate
         then begin canvas.brush.color:=FixedColor;
                    canvas.font.color:=ClBtnText;
                    inherited DrawCell(ACol, ARow, ARect, AState);
              end
         else begin canvas.brush.color:=Color;
                    canvas.font.color:=ClWindowtext;
                    inherited DrawCell(ACol, ARow, ARect, AState);
              end;
      defaultdrawing:=olddefault;
if assigned(FOnDrawCell) then FonDrawCell(self,Acol,Arow,Arect,Astate{,inhalt});
end;

{The following procedures/functions are just copied from TCustomGrid}

procedure TmStrGrid.CalcDrawInfoXY(var DrawInfo: TGridDrawInfo;
  UseWidth, UseHeight: Integer);
  procedure CalcAxis(var AxisInfo: TGridAxisDrawInfo; UseExtent: Integer);
  var
    I: Integer;
  begin
    with AxisInfo do
    begin
      GridExtent := UseExtent;
      GridBoundary := FixedBoundary;
      FullVisBoundary := FixedBoundary;
      LastFullVisibleCell := FirstGridCell;
      for I := FirstGridCell to GridCellCount - 1 do
      begin
        Inc(GridBoundary, GetExtent(I) + EffectiveLineWidth);
        if GridBoundary > GridExtent + EffectiveLineWidth then
        begin
          GridBoundary := GridExtent;
          Break;
        end;
        LastFullVisibleCell := I;
        FullVisBoundary := GridBoundary;
      end;
    end;
  end;

begin
  CalcFixedInfo(DrawInfo);
  CalcAxis(DrawInfo.Horz, UseWidth);
  CalcAxis(DrawInfo.Vert, UseHeight);
end;

procedure TmStrGrid.CalcDrawInfo(var DrawInfo: TGridDrawInfo);
begin
  CalcDrawInfoXY(DrawInfo, ClientWidth, ClientHeight);
end;


function TmStrGrid.CalcCoordFromPoint(X, Y: Integer;
  const DrawInfo: TGridDrawInfo): TGridCoord;

  function DoCalc(const AxisInfo: TGridAxisDrawInfo; N: Integer): Integer;
  var
    I, Start, Stop: Longint;
    Line: Integer;
  begin
    with AxisInfo do
    begin
      if N < FixedBoundary then
      begin
        Start := 0;
        Stop :=  FixedCellCount - 1;
        Line := 0;
      end
      else
      begin
        Start := FirstGridCell;
        Stop := GridCellCount - 1;
        Line := FixedBoundary;
      end;
      Result := -1;
      for I := Start to Stop do
      begin
        Inc(Line, GetExtent(I) + EffectiveLineWidth);
        if N < Line then
        begin
          Result := I;
          Exit;
        end;
      end;
    end;
  end;

begin
  Result.X := DoCalc(DrawInfo.Horz, X);
  Result.Y := DoCalc(DrawInfo.Vert, Y);
end;

procedure TmStrGrid.CalcFixedInfo(var DrawInfo: TGridDrawInfo);

  procedure CalcFixedAxis(var Axis: TGridAxisDrawInfo; LineOptions: TGridOptions;
    FixedCount, FirstCell, CellCount: Integer; GetExtentFunc: TGetExtentsFunc);
  var
    I: Integer;
  begin
    with Axis do
    begin
      if LineOptions * Options = [] then
        EffectiveLineWidth := 0
      else
        EffectiveLineWidth := GridLineWidth;

      FixedBoundary := 0;
      for I := 0 to FixedCount - 1 do
        Inc(FixedBoundary, GetExtentFunc(I) + EffectiveLineWidth);

      FixedCellCount := FixedCount;
      FirstGridCell := FirstCell;
      GridCellCount := CellCount;
      GetExtent := GetExtentFunc;
    end;
  end;

begin
  CalcFixedAxis(DrawInfo.Horz, [goFixedVertLine, goVertLine], FixedCols,
    LeftCol, ColCount, GetColWidths);
  CalcFixedAxis(DrawInfo.Vert, [goFixedHorzLine, goHorzLine], FixedRows,
    TopRow, RowCount, GetRowHeights);
end;

function TmStrGrid.GetColWidths(Index: Longint): Integer;
begin Result:=ColWidths[index]
end;

function TmStrGrid.GetRowHeights(Index: Longint): Integer;
begin Result:=RowHeights[index]
end;

procedure TmStrGrid.CalcSizingState(X, Y: Integer; var State: TGridState;
  var Index: Longint; var SizingPos, SizingOfs: Integer;
  var FixedInfo: TGridDrawInfo);

  procedure CalcAxisState(const AxisInfo: TGridAxisDrawInfo; Pos: Integer;
    NewState: TGridState);
  var
    I, Line, Back, Range: Integer;
  begin
    with AxisInfo do
    begin
      Line := FixedBoundary;
      Range := EffectiveLineWidth;
      Back := 0;
      if Range < 7 then
      begin
        Range := 7;
        Back := (Range - EffectiveLineWidth) shr 1;
      end;
      for I := FirstGridCell to GridCellCount - 1 do
      begin
        Inc(Line, GetExtent(I));
        if Line > GridExtent then Break;
        if (Pos >= Line - Back) and (Pos <= Line - Back + Range) then
        begin
          State := NewState;
          SizingPos := Line;
          SizingOfs := Line - Pos;
          Index := I;
          Exit;
        end;
        Inc(Line, EffectiveLineWidth);
      end;
      if (Pos >= GridExtent - Back) and (Pos <= GridExtent) then
      begin
        State := NewState;
        SizingPos := GridExtent;
        SizingOfs := GridExtent - Pos;
        Index := I;
      end;
    end;
  end;

var
  EffectiveOptions: TGridOptions;
begin
  State := gsNormal;
  Index := -1;
  EffectiveOptions := Options;
  if csDesigning in ComponentState then
  begin
    Include(EffectiveOptions, goColSizing);
    Include(EffectiveOptions, goRowSizing);
  end;
  if [goColSizing, goRowSizing] * EffectiveOptions <> [] then
    with FixedInfo do
    begin
      Vert.GridExtent := ClientHeight;
      Horz.GridExtent := ClientWidth;
      if (X > Horz.FixedBoundary) and (goColSizing in EffectiveOptions) then
      begin
        if Y >= Vert.FixedBoundary then Exit;
        CalcAxisState(Horz, X, gsColSizing);
      end
      else if (Y > Vert.FixedBoundary) and (goRowSizing in EffectiveOptions) then
      begin
        if X >= Horz.FixedBoundary then Exit;
        CalcAxisState(Vert, Y, gsRowSizing);
      end;
    end;
end;

procedure TmStrGrid.assign(x:TStringGrid);
var r:integer;
begin RowCount:=x.RowCount;
      ColCount:=x.ColCount;
      for r:=0 to RowCount-1 do
      Rows[r]:=x.Rows[r]
end;

end.
